"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
exports.defaultOptions = {
    /**
     * React.lazy() only offically supports default exports, but it works if you do:
     * `React.lazy(() => import('./x').then(m => ({ default: m.Xyz })))`
     * This lets you rewrite all imported components. Once React supports a better way to import
     * named exports this transform will be updated to that.
     */
    onlyRewriteDefaultExports: false,
    /**
     * If the import declaration can't be removed because both a component and another function
     * is imported this lets you skip converting to a lazy component since it'll probably be
     * placed in the same bundle either way. The default is `false` since other imports might
     * be dead code eliminated without this transform knowing about it.
     */
    onlyRewriteIfImportCanBeRemoved: false,
    /**
     * Defaults to creating a PropertyAccessExpression of `React.lazy` but can be any expression you want.
     * Use `Options.createImportDeclaration()` if you need to import something from another module to call.
     */
    createComponentWrapperExpression: function (typescript) {
        return typescript.createPropertyAccess(typescript.createIdentifier('React'), typescript.createIdentifier('lazy'));
    },
    /**
     * Implement this if you want to import another module which contains your wrapper expression.
     * This can be used if you want to replace `React.lazy()` with something else.
     */
    createImportDeclaration: function (typescript, currentFile) {
        return null;
    },
    /**
     * This lets you bail on rewriting imports on a per import declaration basis. If you have a list of
     * files to load sync (from for example https://github.com/avensia-oss/ts-transform-instrument-react-components)
     * you can implement this function to only rewrite imports from those files.
     */
    shouldRewrite: function (importSpecifier, currentFile) {
        return true;
    },
};
function transformer(program, options) {
    if (options === void 0) { options = exports.defaultOptions; }
    options = __assign({}, exports.defaultOptions, options);
    if (!program) {
        throw new Error('No ts.Program was passed to the transformer factory');
    }
    return function (context) { return function (file) {
        return visitSourceFile(file, program, context, options);
    }; };
}
exports.default = transformer;
function visitSourceFile(sourceFile, program, context, options) {
    var imports = sourceFile.statements.filter(function (s) { return ts.isImportDeclaration(s); });
    var hasReactImport = !!imports.find(function (s) { return s.moduleSpecifier.text === 'react'; });
    if (hasReactImport && sourceFile.fileName.indexOf('.d.ts') === -1) {
        // Any import which imports a variable with leading uppercase char
        var potentialComponentImports_1 = imports.filter(function (i) {
            return ((i.importClause &&
                i.importClause.name &&
                i.importClause.name.text[0] === i.importClause.name.text[0].toUpperCase()) ||
                (i.importClause &&
                    i.importClause.namedBindings &&
                    ts.isNamedImports(i.importClause.namedBindings) &&
                    i.importClause.namedBindings.elements.some(function (e) { return e.name.text[0] === e.name.text[0].toUpperCase(); })));
        });
        if (potentialComponentImports_1.length) {
            var transformedSourceFile = ts.visitEachChild(visitNode(sourceFile, potentialComponentImports_1, sourceFile, program, options), function (childNode) { return visitNodeAndChildren(childNode, potentialComponentImports_1, sourceFile, program, context, options); }, context);
            return transformedSourceFile;
        }
    }
    return sourceFile;
}
function visitNodeAndChildren(node, potentialComponentImports, sourceFile, program, context, options) {
    var visitedNode = visitNode(node, potentialComponentImports, sourceFile, program, options);
    var visitedChildNode = ts.visitEachChild(visitedNode, function (childNode) { return visitNodeAndChildren(childNode, potentialComponentImports, sourceFile, program, context, options); }, context);
    return visitedChildNode;
}
function getImportedReactComponents(importDecl, typeChecker, options) {
    var componentNames = [];
    if (importDecl.importClause) {
        if (importDecl.importClause.name) {
            if (isComponentType('default', importDecl, typeChecker)) {
                componentNames.push(importDecl.importClause.name.text);
            }
        }
        if (!options.onlyRewriteDefaultExports &&
            importDecl.importClause.namedBindings &&
            ts.isNamedImports(importDecl.importClause.namedBindings)) {
            for (var _i = 0, _a = importDecl.importClause.namedBindings.elements; _i < _a.length; _i++) {
                var element = _a[_i];
                if (isComponentType((element.propertyName || element.name).text, importDecl, typeChecker)) {
                    componentNames.push(element.name.text);
                }
            }
        }
    }
    return componentNames;
}
function isComponentType(identifier, importDecl, typeChecker) {
    var importSymbol = typeChecker.getSymbolAtLocation(importDecl.moduleSpecifier);
    if (importSymbol) {
        var exports_1 = typeChecker.getExportsOfModule(importSymbol);
        var exportedSymbol = exports_1.find(function (e) { return e.name === identifier; });
        if (exportedSymbol && exportedSymbol.valueDeclaration) {
            if (ts.isClassDeclaration(exportedSymbol.valueDeclaration) && exportedSymbol.valueDeclaration.heritageClauses) {
                var maybeComponentClass = exportedSymbol.valueDeclaration;
                if (maybeComponentClass.heritageClauses) {
                    var baseClass = maybeComponentClass.heritageClauses.find(function (h) { return h.token === ts.SyntaxKind.ExtendsKeyword; });
                    if (baseClass) {
                        var extendsReactComponent = !!baseClass.types.find(function (t) {
                            return ts.isExpressionWithTypeArguments(t) &&
                                ts.isPropertyAccessExpression(t.expression) &&
                                ts.isIdentifier(t.expression.expression) &&
                                t.expression.name.escapedText === 'Component' &&
                                t.expression.expression.escapedText === 'React';
                        });
                        return extendsReactComponent;
                    }
                }
            }
            else {
                var func = getFunction(exportedSymbol.valueDeclaration);
                if (func) {
                    var type = typeChecker.getTypeOfSymbolAtLocation(exportedSymbol, exportedSymbol.valueDeclaration);
                    var callSignatures = type.getCallSignatures();
                    if (callSignatures.length) {
                        var callSignature = callSignatures[0];
                        var returnType = typeChecker.getReturnTypeOfSignature(callSignature).getSymbol();
                        var returnParentType = returnType ? returnType.parent : null;
                        if (
                        // props and context
                        (func.parameters.length === 1 || func.parameters.length === 2) &&
                            returnType &&
                            returnType.escapedName === 'Element' &&
                            returnParentType &&
                            returnParentType.escapedName === 'JSX') {
                            return true;
                        }
                    }
                }
            }
        }
    }
    return false;
}
function getFunction(decl) {
    if (ts.isFunctionDeclaration(decl)) {
        return decl;
    }
    if (ts.isExportAssignment(decl)) {
        if (ts.isFunctionExpression(decl.expression) || ts.isArrowFunction(decl.expression)) {
            return decl.expression;
        }
        return null;
    }
    if (ts.isVariableStatement(decl) || ts.isVariableDeclaration(decl)) {
        var declaration = ts.isVariableStatement(decl) ? decl.declarationList.declarations[0] : decl;
        if (declaration.initializer &&
            (ts.isArrowFunction(declaration.initializer) || ts.isFunctionExpression(declaration.initializer))) {
            return declaration.initializer;
        }
    }
    return null;
}
function visitNode(node, potentialComponentImports, sourceFile, program, options) {
    if (ts.isImportDeclaration(node) && potentialComponentImports.indexOf(node) !== -1 && node.importClause) {
        if (!options.shouldRewrite(node.moduleSpecifier.text, sourceFile.fileName)) {
            return node;
        }
        var defaultImportName_1 = node.importClause.name ? node.importClause.name.escapedText.toString() : undefined;
        var componentNames = getImportedReactComponents(node, program.getTypeChecker(), options);
        if (componentNames.length) {
            var defaultName = defaultImportName_1 && componentNames.indexOf(defaultImportName_1) !== -1 ? undefined : defaultImportName_1;
            var additionalStatements = componentNames.map(function (c) {
                var importCall = ts.createCall({ kind: ts.SyntaxKind.ImportKeyword, flags: ts.NodeFlags.None }, undefined, [node.moduleSpecifier]);
                var arrowBody = importCall;
                if (defaultImportName_1 !== c) {
                    var element = node.importClause.namedBindings.elements.find(function (e) { return e.name.escapedText == c; });
                    if (element) {
                        arrowBody = ts.createCall(ts.createPropertyAccess(importCall, 'then'), undefined, [
                            ts.createArrowFunction(undefined, undefined, [ts.createParameter(undefined, undefined, undefined, 'm', undefined, undefined, undefined)], undefined, undefined, ts.createParen(ts.createObjectLiteral([
                                ts.createPropertyAssignment('default', ts.createPropertyAccess(ts.createIdentifier('m'), (element.propertyName || element.name).escapedText.toString())),
                            ]))),
                        ]);
                    }
                }
                return ts.createVariableStatement(undefined, ts.createVariableDeclarationList([
                    ts.createVariableDeclaration(c, undefined, ts.createCall(options.createComponentWrapperExpression(ts), undefined, [
                        ts.createArrowFunction(undefined, undefined, [], undefined, undefined, arrowBody),
                    ])),
                ], ts.NodeFlags.Const));
            });
            var additionalImport = options.createImportDeclaration(ts, sourceFile.fileName);
            if (additionalImport) {
                additionalStatements.unshift(additionalImport);
            }
            var importedNames = (node.importClause.name && componentNames.indexOf(node.importClause.name.escapedText.toString()) !== -1
                ? 1
                : 0) +
                (node.importClause.namedBindings ? node.importClause.namedBindings.elements.length : 0);
            if (importedNames === componentNames.length) {
                return additionalStatements;
            }
            else if (!options.onlyRewriteIfImportCanBeRemoved) {
                return additionalStatements.concat([
                    ts.createImportDeclaration(node.decorators, node.modifiers, ts.createImportClause(defaultName ? ts.createIdentifier(defaultName) : undefined, node.importClause.namedBindings
                        ? removeImportNames(node.importClause.namedBindings, componentNames)
                        : undefined), node.moduleSpecifier),
                ]);
            }
        }
    }
    return node;
}
function removeImportNames(namedBindings, importNamesToRemove) {
    return __assign({}, namedBindings, { elements: namedBindings.elements.filter(function (e) { return importNamesToRemove.indexOf((e.propertyName || e.name).text) === -1; }) });
}
